/*
 * Created on Jun 23, 2003
 *
 */
package luke;

import java.awt.Font;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;

import javax.swing.JFileChooser;
import javax.swing.UIManager;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.index.TermEnum;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.store.FSDirectory;

import thinlet.FrameLauncher;
import thinlet.Thinlet;

/**
 * This class allows you to browse a <a href="jakarta.apache.org/lucene">Lucene</a>
 * index in several ways - by document, by term, by query, and by most frequent terms.
 * 
 * @author Andrzej Bialecki <ab@getopt.org>
 *
 */
public class Luke extends Thinlet implements ClipboardOwner {

	private PersistentStringList mruList = new PersistentStringList( new File( HOME_DIR+SEP_CHAR + PROP_NAME), "mru", 10 );	    
    private FSDirectory dir = null;
    private String pName = null;
    private IndexReader ir = null;
    private Collection fn = null;
    private String[] idxFields = null;
    private Object error = null;
    private Object statmsg = null;
    private Analyzer stdAnalyzer = new StandardAnalyzer();
    private Analyzer analyzer = null;
    private QueryParser qp = null;
    private boolean readOnly = false;
    private boolean useCompound = false;
    
    private static final String MSG_NOINDEX = "FAILED: No index, or index is closed. Reopen it.";
    private static final String MSG_READONLY = "FAILED: Read-Only index.";
	private static final String PROP_NAME = ".luke";
	private static final String HOME_DIR = System.getProperty("user.home");
	private static final String SEP_CHAR = System.getProperty("file.separator");
	
    public Luke() {
        super();
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {};
        setColors(0xece9d8, 0x000000, 0xf5f4f0, 0x919b9a, 0xb0b0b0, 0xededed, 0xb9b9b9, 0xff899a, 0xc5c5dd);
        addComponent(this, "luke.xml", null, null);
        error = addComponent(null, "error.xml", null, null);
        statmsg = find("statmsg");
        actionOpen();
    }
    
    private void showStatus(final String msg) {
        Thread thr = new Thread() {
            public void run() {
                setString(statmsg, "text", msg);
                try {
                    sleep(5000);
                } catch (Exception e) {};
                setString(statmsg, "text", "");
            }
        };
        thr.start();
    }
    
    public Object addComponent(Object parent, String compView, String handlerStr, Object[] argv) {
        Object res = null;
        Object handler = null;
        try {
            if (handlerStr != null) {
                if (argv == null) {
                    handler = Class.forName(handlerStr)
                        .getConstructor(new Class[] { Thinlet.class })
                        .newInstance(new Object[] { this });
                } else {
                    handler = Class.forName(handlerStr)
                        .getConstructor(new Class[] { Thinlet.class, Object[].class})
                        .newInstance(new Object[] { this, argv});
                }
            }
            if (handler != null) {
                res = parse(compView, handler);
            } else
                res = parse(compView);
            if (parent != null) {
                if (parent instanceof Thinlet)
                    add(res);
                else
                    add(parent, res);
            }
            return res;
        } catch (Exception exc) {
            exc.printStackTrace();
            errorMsg(exc.getMessage());
            return null;
        }
    }

    public void errorMsg(String msg) {
        System.out.println(msg);
        Object fMsg = find(error, "msg");
        setString(fMsg, "text", msg);
        add(error);
    }
    
    public void actionOpen() {
        addComponent(this, "lukeinit.xml", null, null);
    }
    
    public void openBrowse(Object path) {
        JFileChooser fd = new JFileChooser();
        fd.setDialogType(JFileChooser.OPEN_DIALOG);
        fd.setDialogTitle("Select Index directory");
        fd.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        fd.setFileHidingEnabled(false);
        int res = fd.showOpenDialog(this);
        File iDir = null;
        if (res == JFileChooser.APPROVE_OPTION) iDir = fd.getSelectedFile();
        if (iDir != null && iDir.exists()) {
            if (!iDir.isDirectory()) iDir = iDir.getParentFile();
            setString(path, "text", iDir.toString());
        }
    }
	public void setupInit(Object dialog) {
		Object path = find(dialog, "path");
		syncMRU(path);
	}
	
    public void openOk(Object dialog) {
        Object path = find(dialog, "path");
        pName = getString(path, "text").trim();
                
        boolean force = getBoolean(find(dialog, "force"), "selected");
        if (pName == null || pName.trim().equals("") || !IndexReader.indexExists(pName)) {
            errorMsg("Invalid path, or not a Lucene index.");
            return;
        }
        readOnly = getBoolean(find(dialog, "ro"), "selected");
        remove(dialog);
        if (dir != null) {
            try {
                if (ir != null) ir.close();
            } catch(Exception e) {};
            try {
                dir.close();
            } catch(Exception e) {};
        }
        try {
        	mruList.add(pName);
			syncMRU(path);
        	        	
            dir = FSDirectory.getDirectory(pName, false);
            if (IndexReader.isLocked(dir)) {
                if (readOnly) {
                    showStatus("Index is locked and Read-Only. Open for read-write and 'Force unlock'.");
                    dir.close();
                    return;
                }
                if (force) {
                    IndexReader.unlock(dir);
                } else {
                    showStatus("Index is locked. Try 'Force unlock' when opening.");
                    dir.close();
                    return;
                }
            }
            showFiles(dir);
			Object fileSize = find("iFileSize");
			BigInteger totalFileSize = calcTotalFileSize(pName, dir);
			setString(fileSize, "text", normalizeSize(totalFileSize) + normalizeUnit(totalFileSize));

            System.out.println(dir);
            ir = IndexReader.open(dir);
            Object pOver = find("pOver");
            Object iName = find("idx");
            setString(iName, "text", pName + (readOnly? " (R)" : ""));
            iName = find(pOver, "iName");
            setString(iName, "text", pName + (readOnly? " (Read-Only)" : ""));
            Object iDocs = find(pOver, "iDocs");
            String numdocs = String.valueOf(ir.numDocs());
            setString(iDocs, "text", numdocs);
            iDocs = find("iDocs1");
            setString(iDocs, "text", String.valueOf(ir.numDocs() - 1));
            Object iFields = find(pOver, "iFields");
            fn = ir.getFieldNames();
            idxFields = new String[fn.size()];
            setString(iFields, "text", String.valueOf(fn.size()));
            Object fList = find(pOver, "fList");
            Object defFld = find("defFld");
            Object fCombo = find("fCombo");
            java.util.Iterator it = fn.iterator();
            int i = 0;
			removeAll(fList);

            while (it.hasNext()) {
                Object item = create("item");
                add(fList, item);
                String name = (String)it.next();
                setString(item, "text", "<" + name + ">");
                putProperty(item, "fName", name);
                item = create("choice");
                add(defFld, item);
                setString(item, "text", name);
                item = create("choice");
                add(fCombo, item);
                setString(item, "text", name);
                putProperty(item, "fName", name);
                idxFields[i++] = name;
            }
            setString(defFld, "text", idxFields[0]);
            Object iTerms = find(pOver, "iTerms");
            TermEnum te = ir.terms();
            int cnt = 0;
            while (te.next()) cnt++;
            te.close();
            setString(iTerms, "text", String.valueOf(cnt));
            Object iMod = find(pOver, "iMod");
            setString(iMod, "text", new java.util.Date(IndexReader.lastModified(dir)).toString());
            setString(find("defFld"), "text", idxFields[0]);
            // Remove columns
            Object header = get(find("sTable"), "header");
            Object col = get(header, ":comp");  // rank
            col = get(col, ":next");            // id
            Object tail = col;
            Object next = get(col, ":next");
            while (next != null) {
                set (col, ":next", null);
                col = next;
                next = get(col, ":next");
            }
            for (int j = 0; j < idxFields.length; j++) {
                Object c = create("column");
                setString(c, "text", idxFields[j]);
                set(tail, ":next", c);
                tail = c;
            }
            actionTopTerms(find("nTerms"));
            showStatus("Index successfully open.");
        } catch (Exception e) {
            e.printStackTrace();
            errorMsg(e.getMessage());
            return;
        }
    }

    public void showFiles(FSDirectory dir) throws Exception {
        String[] files = dir.list();
        Object filesTable = find("filesTable");
        removeAll( filesTable );
        for(int i = 0; i < files.length; i++) {
            String filename;
            if(pName.endsWith(SEP_CHAR)) {
                filename = pName;                   
            } else {
                filename = pName + SEP_CHAR;                    
            }       
            File file = new File(filename + files[i]);
            Object row = create("row");
            Object nameCell = create("cell");
            setString(nameCell, "text", files[i]);
            add(row, nameCell);
            Object sizeCell = create("cell");
            setString(sizeCell, "text", normalizeSize(file.length()));
            add(row, sizeCell);
            Object unitCell = create("cell");
            setString(unitCell, "text", normalizeUnit(file.length()));                  
            add(row, unitCell);
            add(filesTable, row);
        }            
    }

	private void syncMRU(Object path) {
		removeAll(path);
		for (Iterator iter = mruList.iterator(); iter.hasNext(); ) {
			String element = (String) iter.next();
			Object choice = create("choice");
			setString(choice, "text", element);
			add(path, choice);				
		}
	}

	private String normalizeUnit(long len) {
		if(len == 1) {
			return " byte";
		} else if(len < 1024) {
			return " bytes";			
		} else if(len < 51200000) {
			return " kB";
		} else {
			return " MB";
		}
	}

	private String normalizeUnit(BigInteger len) {
		if(len.compareTo(new BigInteger("1")) == 0)  {
			return " byte";
		} else if(len.compareTo(new BigInteger("1024")) == -1) {
			return " bytes";			
		} else if(len.compareTo(new BigInteger("51200000")) == -1) {
			return " kB";
		} else {
			return " MB";
		}
	}
    
	private String normalizeSize(BigInteger len) {
		if(len.compareTo(new BigInteger("1")) == 0)  {
			return len.toString();
		} else if(len.compareTo(new BigInteger("1024")) == -1) {
			return len.toString();			
		} else if(len.compareTo(new BigInteger("51200000")) == -1) {
			return len.divide(new BigInteger("1024")).toString();
		} else {
			return len.divide(new BigInteger("102400")).toString();
		}
	}

	private BigInteger calcTotalFileSize(String path, FSDirectory fsdir) {
		BigInteger totalFileSize = new BigInteger("0");
		try {
			String[] files;
			files = fsdir.list();
			for(int i = 0; i < files.length; i++) {
				String filename;
		        if(pName.endsWith(SEP_CHAR)) {
					filename = pName;        			
        		} else {
					filename = pName + SEP_CHAR;					
				}		

				File file = new File(filename + files[i]);
				totalFileSize = totalFileSize.add(new BigInteger(file.length() + ""));
			}
		} catch (IOException e) {
			// swallow .. degrade gracefully
		}
		            
		return totalFileSize;
		
	}    
	private String normalizeSize(long len) {
		if(len == 1) {
			return len + "";
		} else if(len < 1024) {
			return len + "";			
		} else if(len < 51200000) {
			return (len / 1024) + "";
		} else {
			return (len / 102400) + "";
		}
	}



	public void actionTopTerms(Object nTerms) {
        int ndoc = getInteger(nTerms, "value");
        System.out.println("nTerms=" + ndoc);
        Object[] fields = getSelectedItems(find("fList"));
        String[] flds = null;
        if (fields == null || fields.length == 0) {
            flds = idxFields;
        } else {
            flds = new String[fields.length];
            for (int i = 0; i < fields.length; i++) {
                flds[i] = (String)getProperty(fields[i], "fName");
                System.out.println(" - " + flds[i]);
            }
        }
        try {
            TermInfo[] tis = HighFreqTerms.getHighFreqTerms(dir, null, ndoc, flds);
            Object table = find("tTable");
            removeAll(table);
            if (tis == null || tis.length == 0) {
                System.out.println("tis=" + tis + " (no results)");
                Object row = create("row");
                Object cell = create("cell");
                add(row, cell);
                cell = create("cell");
                add(row, cell);
                cell = create("cell");
                add(row, cell);
                cell = create("cell");
                setBoolean(cell, "enabled", false);
                setString(cell, "text", "No Results");
                add(row, cell);
                add(table, row);
                return;
            } 
            for (int i = 0; i < tis.length; i++) {
                Object row = create("row");
                add(table, row);
                putProperty(row, "term", tis[i].term);
                Object cell = create("cell");
                setChoice(cell, "alignment", "right");
                setString(cell, "text", String.valueOf(i + 1));
                add(row, cell);
                cell = create("cell");
                setChoice(cell, "alignment", "right");
                setString(cell, "text", String.valueOf(tis[i].docFreq) + "  ");
                add(row, cell);
                cell = create("cell");
                setString(cell, "text", " <" + tis[i].term.field() + "> ");
                add(row, cell);
                cell = create("cell");
                setString(cell, "text", "  " + tis[i].term.text());
                add(row, cell);
            }
        } catch(Exception e) {
            e.printStackTrace();
            errorMsg(e.getMessage());
        }
    }
    
    public void browseTermDocs(Object tTable) {
        Object row = getSelectedItem(tTable);
        if (row == null) return;
        Term t = (Term)getProperty(row, "term");
        if (t == null) return;
        Object tabpane = find("maintpane");
        setInteger(tabpane, "selected", 1);
        _showTerm(find("fCombo"), find("fText"), t);
        repaint();        
        
    }
    
    public void showTermDocs(Object tTable) {
        Object row = getSelectedItem(tTable);
        if (row == null) return;
        Term t = (Term)getProperty(row, "term");
        if (t == null) return;
        Object tabpane = find("maintpane");
        setInteger(tabpane, "selected", 2);
        Object qField = find("qField");
        setString(qField, "text", t.field() + ":" + t.text());
        search(qField);
        repaint();        
        
    }
    
    public void actionUseCompound(Object check) {
        useCompound = getBoolean(check, "selected");
    }
    
    public void actionOptimize() {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        if (readOnly) {
            showStatus(MSG_READONLY);
            return;
        }
        try {
            ir.close();
            IndexWriter iw = new IndexWriter(dir, new org.apache.lucene.analysis.WhitespaceAnalyzer(), false);
            iw.setUseCompoundFile(useCompound);
            BigInteger startSize = calcTotalFileSize(pName, dir);
            long startTime = System.currentTimeMillis();
            iw.optimize();
			long endTime = System.currentTimeMillis();            
			BigInteger endSize = calcTotalFileSize(pName, dir);
			BigInteger deltaSize = startSize.subtract( endSize );
			String sign = deltaSize.compareTo(new BigInteger("0")) < 0 ? " Increased " : " Reduced ";
			String sizeMsg = sign + normalizeSize(deltaSize.abs()) + normalizeUnit(deltaSize.abs());
            String timeMsg = String.valueOf(endTime - startTime) + " ms";
            showStatus(sizeMsg + " in " + timeMsg);
            showFiles(dir);
            iw.close();
            ir = IndexReader.open(dir);
                       
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public void showPrevDoc(Object docNum) {
        _showDoc(docNum, -1);
    }
    
    public void showNextDoc(Object docNum) {
        _showDoc(docNum, +1);
    }
    
    public void showDoc(Object docNum) {
        _showDoc(docNum, 0);
    }
    
    private void _showDoc(Object docNum, int incr) {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        String num = getString(docNum, "text");
        if (num.trim().equals("")) num = String.valueOf(-incr);
        try {
            int iNum = Integer.parseInt(num);
            iNum += incr;
            if (iNum < 0 || iNum >= ir.numDocs()) {
                showStatus("Document number outside valid range.");
                return;
            }
            Document doc = null;
            if (!ir.isDeleted(iNum)) doc = ir.document(iNum);
            else showStatus("This document has been deleted and is not available.");
            _showDocFields(iNum, doc);
        } catch(Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    private void _showDocFields(int docid, Document doc) {
        Object table = find("docTable");
        setString(find("docNum"), "text", String.valueOf(docid));
        setString(find("docNum1"), "text", String.valueOf(docid));
        removeAll(table);
        putProperty(table, "doc", doc);
        if (doc == null) return;
        for (Enumeration e = doc.fields(); e.hasMoreElements(); ) {
            Field f = (Field)e.nextElement();
            Object row = create("row");
            add(table, row);
            putProperty(row, "field", f);
            Object cell = create("cell");
            setString(cell, "text", "<" + f.name() + ">");
            add(row, cell);
            cell = create("cell");
            if (f.isIndexed()) setString(cell, "text", "+");
            add(row, cell);
            cell = create("cell");
            if (f.isTokenized()) setString(cell, "text", "+");
            add(row, cell);
            cell = create("cell");
            if (f.isStored()) setString(cell, "text", "+");
            add(row, cell);
            cell = create("cell");
            setString(cell, "text", f.stringValue());
            add(row, cell);
        }
    }
    
    public void clipCopyFields(Object table) {
        Object[] rows = getSelectedItems(table);
        if (rows == null || rows.length == 0) return;
        Document doc = (Document)getProperty(table, "doc");
        if (doc == null) return;
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < rows.length; i++) {
            if (i > 0) sb.append('\n');
            sb.append(((Field)getProperty(rows[i], "field")).toString());
        }
        StringSelection sel = new StringSelection(sb.toString());
        Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel, this);
    }
    
    public void clipCopyDoc(Object table) {
        Document doc = (Document)getProperty(table, "doc");
        if (doc == null) return;
        StringBuffer sb = new StringBuffer();
        Object[] rows = getItems(table);
        for (int i = 0; i < rows.length; i++) {
            if (i > 0) sb.append('\n');
            sb.append(((Field)getProperty(rows[i], "field")).toString());
        }
        StringSelection sel = new StringSelection(sb.toString());
        Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel, this);
    }
    
    public void showFirstTerm(Object fCombo, Object fText) {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        try {
            TermEnum te = ir.terms();
            te.next();
            Term t = te.term();
            _showTerm(fCombo, fText, t);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public void showNextTerm(Object fCombo, Object fText) {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        try {
            String text = getString(fText, "text");
            String fld = getString(fCombo, "text");
            TermEnum te = null;
            if (text == null || text.trim().equals("")) te = ir.terms();
            else te = ir.terms(new Term(fld, text));
            te.next();
            Term t = te.term();
            _showTerm(fCombo, fText, t);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public void showTerm(Object fCombo, Object fText) {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        try {
            String text = getString(fText, "text");
            String fld = getString(fCombo, "text");
            if (text == null || text.trim().equals("")) return;
            Term t = new Term(fld, text);
            _showTerm(fCombo, fText, t);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }

    private void _showTerm(Object fCombo, Object fText, Term t) {    
        if (t == null) {
            showStatus("No terms?!");
            return;            
        }
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        Object[] choices = getItems(fCombo);
        for (int i = 0; i < choices.length; i++) {
            if (t.field().equals(getString(choices[i], "text"))) {
                setInteger(fCombo, "selected", i);
                break;
            }
        }
        setString(fText, "text", t.text());
        putProperty(fText, "term", t);
        putProperty(fText, "td", null);
        setString(find("tdNum"), "text", "?");
        setString(find("tFreq"), "text", "?");
        Object dFreq = find("dFreq");
        try {
            int freq = ir.docFreq(t);
            setString(dFreq, "text", String.valueOf(freq));
            dFreq = find("tdMax");
            setString(dFreq, "text", String.valueOf(freq));
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
            setString(dFreq, "text", "?");
        }
    }
    
    public void showFirstTermDoc(Object fText) {
        Term t = (Term)getProperty(fText, "term");
        if (t == null) return;
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        try {
            TermDocs td = ir.termDocs(t);
            td.next();
            setString(find("tdNum"), "text", "1");
            putProperty(fText, "td", td);
            _showTermDoc(fText, td);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public void showNextTermDoc(Object fText) {
        Term t = (Term)getProperty(fText, "term");
        if (t == null) return;
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        try {
            TermDocs td = (TermDocs)getProperty(fText, "td");
            if (td == null) {
                showFirstTermDoc(fText);
                return;
            }
            if (!td.next()) return;
            Object tdNum = find("tdNum");
            String sCnt = getString(tdNum, "text");
            int cnt = 1;
            try {
                cnt = Integer.parseInt(sCnt);
            } catch(Exception e) {};
            setString(tdNum, "text", String.valueOf(cnt + 1));
            _showTermDoc(fText, td);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public void showAllTermDoc(Object fText) {
        Term t = (Term)getProperty(fText, "term");
        if (t == null) return;
        if (ir == null) {
            showStatus("MSG_NOINDEX");
            return;
        }
        Object tabpane = find("maintpane");
        setInteger(tabpane, "selected", 2);
        Object qField = find("qField");
        setString(qField, "text", t.field() + ":" + t.text());
        showParsed();
        IndexSearcher is = null;
        try {
	        is = new IndexSearcher(dir);
	        Object sTable = find("sTable");
	        removeAll(sTable);
	        _search(new TermQuery(t), is, sTable);
        } catch (Exception e) {
        	e.printStackTrace();
        	errorMsg(e.getMessage());
        } finally {
        	if (is != null) try {is.close();} catch (Exception e1) {};
        }
    }

    private QueryParser createQueryParser() {
    	String sAnal = getString(find("cbType"), "text");
    	if (sAnal.trim().equals("")) {
    		sAnal = "org.apache.lucene.analysis.standard.StandardAnalyzer";
    		setString(find("cbType"), "text", sAnal);
    	}
    	try {
    		analyzer = (Analyzer)Class.forName(sAnal).getConstructor(new Class[0]).newInstance(new Object[0]);
    	} catch (Exception e) {
    		e.printStackTrace();
    		errorMsg("Analyzer '" + sAnal + "' error: " + e.getMessage() + ". Using StandardAnalyzer.");
    		analyzer = stdAnalyzer;
    	}
    	String defField = getString(find("defFld"), "text");
    	if (defField == null || defField.trim().equals("")) {
    		defField = idxFields[0];
    		setString(find("defFld"), "text", defField);
    	}
    	return new QueryParser(defField, analyzer);
    }

    public void showParsed() {
    	QueryParser qp = createQueryParser();
    	Object qField = find("qField");
    	Object qFieldParsed = find("qFieldParsed");
    	String queryS = getString(qField, "text");
    	if (queryS.trim().equals("")) {
    		setBoolean(qFieldParsed, "enabled", false);
    		setString(qFieldParsed, "text", "<Empty query>");
    		return;
    	}
    	setBoolean(qFieldParsed, "enabled", true);
    	try {
    		Query q = qp.parse(queryS);
    		setString(qFieldParsed, "text", q.toString());
    	} catch (Throwable t) {
    		setString(qFieldParsed, "text", t.getMessage());
    	}
    }

    public void search(Object qField) {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        qp = createQueryParser();
        Object sTable = find("sTable");
        removeAll(sTable);
        IndexSearcher is = null;
        try {
            String queryS = getString(qField, "text");
            if (queryS.trim().equals("")) {
                showStatus("FAILED: Empty query.");
                return;
            }
            Query q = qp.parse(queryS);
            is = new IndexSearcher(dir);
            putProperty(qField, "searcher", is);
            _search(q, is, sTable);
        } catch (Throwable e) {
        	e.printStackTrace();
        	errorMsg(e.getMessage());
        } finally {
        	if (is != null) try {is.close();} catch (Exception e1) {};
        }
    }
    
    private void _search(Query q, IndexSearcher is, Object sTable) throws Exception {
        long startTime = System.currentTimeMillis();
        Hits hits = is.search(q);
		long endTime = System.currentTimeMillis();
		showStatus( endTime - startTime + " ms");
        if (hits == null || hits.length() == 0) {
            Object row = create("row");
            Object cell = create("cell");
            add(sTable, row);
            add(row, cell);
            cell = create("cell");
            add(row, cell);
            cell = create("cell");
            setString(cell, "text", "No Results");
            setBoolean(cell, "enabled", false);
            add(row, cell);
            setString(find("resNum"), "text", "0");
            return;
        }

        setString(find("resNum"), "text", String.valueOf(hits.length()));
        for (int i = 0; i < hits.length(); i++) {
            Object row = create("row");
            Object cell = create("cell");
            add(sTable, row);
            setString(cell, "text", String.valueOf((double)Math.round((float)1000 * hits.score(i)) / 10.0));
            add(row, cell);
            cell = create("cell");
            setString(cell, "text", String.valueOf(hits.id(i)));
            add(row, cell);
            Document doc = hits.doc(i);
            putProperty(row, "docid", new Integer(hits.id(i)));
            for (int j = 0; j < idxFields.length; j++) {
                cell = create("cell");
                setString(cell, "text", doc.get(idxFields[j]));
                add(row, cell);
            }
        }
        putProperty(sTable, "query", q);
    }
    
    public void explainResult(Object sTable) {
    	Object row = getSelectedItem(sTable);
    	if (row == null) return;
    	Integer docid = (Integer)getProperty(row, "docid");
    	if (docid == null) return;
    	if (ir == null) {
    		showStatus(MSG_NOINDEX);
    		return;
    	}
    	Query q = (Query)getProperty(sTable, "query");
    	if (q == null) return;
    	try {
    		IndexSearcher is = new IndexSearcher(dir);
    		Explanation expl = is.explain(q, docid.intValue());
    		Object dialog = addComponent(null, "explain.xml", null, null);
    		Object eTree = find(dialog, "eTree");
    		addNode(eTree, expl);
    		//setBoolean(eTree, "expand", true);
    		add(dialog);
    	} catch (Exception e) {
    		e.printStackTrace();
    		errorMsg(e.getMessage());
    	}
    }
    
    private void addNode(Object tree, Explanation expl) {
    	Object node = create("node");
    	setString(node, "text", expl.getValue() + "  " + expl.getDescription());
    	add(tree, node);
    	if (getClass(tree) == "tree") {
    		setFont(node, getFont().deriveFont(Font.BOLD));
    	}
    	Explanation[] kids = expl.getDetails();
    	if (kids != null && kids.length > 0) {
    		for (int i = 0; i < kids.length; i++) {
    			addNode(node, kids[i]);
    		}
    	}
    }
    
    public void gotoDoc(Object sTable) {
        System.out.println("gotoDoc");
        Object row = getSelectedItem(sTable);
        if (row == null) return;
        Integer docid = (Integer)getProperty(row, "docid");
        if (docid == null) return;
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        Document doc = null;
        try {
            doc = ir.document(docid.intValue());
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
            return;
        }
        _showDocFields(docid.intValue(), doc);
        Object tabpane = find("maintpane");
        setInteger(tabpane, "selected", 1);
        repaint();
    }
    
    private void _showTermDoc(Object fText, TermDocs td) {
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        try {
            Document doc = ir.document(td.doc());
            setString(find("docNum"), "text", String.valueOf(td.doc()));
            setString(find("tFreq"), "text", String.valueOf(td.freq()));
            _showDocFields(td.doc(), doc);                
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public void deleteTermDoc(Object fText) {
        Term t = (Term)getProperty(fText, "term");
        if (t == null) return;
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        if (readOnly) {
            showStatus(MSG_READONLY);
            return;
        }
        try {
            showNextTerm(find("fCombo"), fText);
            ir.delete(t);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
        
    }

    public void deleteDoc(Object docNum) {
        int docid = 0;
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        if (readOnly) {
            showStatus(MSG_READONLY);
            return;
        }
        try {
            docid = Integer.parseInt(getString(docNum, "text"));
            showNextDoc(docNum);
            ir.delete(docid);
        } catch (Exception e) {
            showStatus(e.getMessage());
            e.printStackTrace();
        }
        
    }

    public void deleteDocList(Object sTable) {
        Object[] rows = getSelectedItems(sTable);
        if (rows == null || rows.length == 0) return;
        if (ir == null) {
            showStatus(MSG_NOINDEX);
            return;
        }
        if (readOnly) {
            showStatus(MSG_READONLY);
            return;
        }
        for (int i = 0; i < rows.length; i++) {
            Integer docId = (Integer)getProperty(rows[i], "docid");
            if (docId == null) continue;
            
        }
    }

    public void actionAbout() {
        Object about = addComponent(this, "about.xml", null, null);
        requestFocus(find(about, "bOk"));
    }
    
    public boolean destroy() {
        if (ir != null) try {ir.close();} catch (Exception e) {};
        if (dir != null) try {dir.close();} catch (Exception e) {};
        return super.destroy();        
    }
    
    public void actionExit() {
        destroy();
        System.exit(0);
    }
    
    public void goUrl(Object url) {
        String u = (String)getProperty(url, "url");
        if (u == null) return;
        try {
            BrowserLauncher.openURL(u);
        } catch (Exception e) {
            e.printStackTrace();
            showStatus(e.getMessage());
        }
    }
    
    public static void main(String[] args) {
        FrameLauncher f = new FrameLauncher("Luke - Lucene Index Browser, v 0.4 (2004-01-10)", new Luke(), 640, 400);
        f.setIconImage(Toolkit.getDefaultToolkit().createImage(Luke.class.getResource("luke.gif")));
        f.start();
    }
    
    /* (non-Javadoc)
     * @see java.awt.datatransfer.ClipboardOwner#lostOwnership(java.awt.datatransfer.Clipboard, java.awt.datatransfer.Transferable)
     */
    public void lostOwnership(Clipboard arg0, Transferable arg1) {

    }

}
